#include <windows.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "world.h"
#include "funcs.h"
int get_empty_obj_index(World *w)
{
    int i;
    for(i = 0; i<1000; i++)
    {
        if(w->objs[i].exist == 0)
            return i;
    }
    return -1;
}

Object mk_dummy(World *w, int x, int y)
{
    Object o;
    strcpy(o.shape.str,"d");
    o.shape.bg_color = 0;
    o.shape.fg_color = 15;
    o.shape.layer = 1;
    o.exist = 1;
    o.x = x;
    o.y = y;
    o.tick_fn = 0;
    o.overlap_fn = 0;

    o.world = w;
    return o;
}

Object mk_bomb(World *w, int x, int y)
{
    Object o = mk_dummy(w,x,y);
    strcpy(o.shape.str,"!");
    o.shape.fg_color = 14;
    o.tick_fn = fn_bomb_timer;
    o.timer = 100;
    return o;
}
Object mk_fire(World *w, int x, int y)
{
    Object o = mk_dummy(w,x,y);
    strcpy(o.shape.str,"*");
    o.shape.fg_color = 13;
    o.tick_fn = fn_fire_timer;
    o.overlap_fn = fn_kill;
    o.timer = 50;
    return o;
}
Object mk_enemy(World * w, int x,int y)
{
    Object o = mk_dummy(w,x,y);
    strcpy(o.shape.str,"e");
    o.shape.fg_color = 12;
    o.tick_fn = fn_do_rand;
    return o;
}
Object mk_hero(World *w, int x, int y)
{
    Object o = mk_dummy(w,x,y);
    strcpy(o.shape.str,"@");
    o.shape.fg_color = 15;
    return o;
}



int new_obj(World * w, char * type, int x, int y)
{
    Object o;
    if(!strcmp(type,"bomb"))
    {
        o = mk_bomb(w,x,y);
    }
    else if(!strcmp(type,"fire"))
    {
        o = mk_fire(w,x,y);
    }
    else if(!strcmp(type,"enemy"))
    {
        o = mk_enemy(w,x,y);
    }
    else if(!strcmp(type,"hero"))
    {
        o = mk_hero(w,x,y);
    }
    else
        return 0;// false
    int k = get_empty_obj_index(w);
    w->objs[k] = o;
    if(!strcmp(type,"hero"))
    {
        w->hero = &(w->objs[k]);
    }
    return 1;
}
int check_block(World * w, int x, int y)
{
    int i;
    if(x >= 100|| y >= 100 || x <0||y<0)
        return 1;
    if(w->map[y][x].layer == 1)
        return 1;

    for(i= 0; i<1000; i++)
    {
        Object * o = &(w->objs[i]);
        if(o->exist == 0)
            continue;
        if(o->x == x &&
           o->y == y &&
           o->shape.layer == 1)
        {
            return 1;
        }
    }
    return 0;
}
void obj_move(World * w, Object * o, int ox, int oy)
{
    int tox = o->x + ox;
    int toy = o->y + oy;
    if(check_block(w,tox,toy) == 0)
    {
        o->x = tox;
        o->y = toy;
    }

}
void key_process(World * w)
{

#if 1
    if(GetAsyncKeyState('W'))
        obj_move(w, w->hero, 0,-1);
    if(GetAsyncKeyState('S'))
        obj_move(w, w->hero, 0,1);
    if(GetAsyncKeyState('A'))
        obj_move(w,w->hero,-1,0);
    if(GetAsyncKeyState('D'))
        obj_move(w,w->hero, 1,0);
    if(GetAsyncKeyState(' '))
    {
        int hx = w->hero->x;
        int hy = w->hero->y;
        new_obj(w, "bomb",hx,hy);
    }
#endif


#if 0
    if(kbhit())
    {
        int k = getch();

        if(k == 'w')
            obj_move(w, w->hero, 0,-1);
        if(k == 's')
            obj_move(w, w->hero, 0,1);
        if(k == 'a')
            obj_move(w,w->hero,-1,0);
        if(k == 'd')
            obj_move(w,w->hero, 1,0);
        if(k == ' ')
        {
            int hx = w->hero->x;
            int hy = w->hero->y;
            new_obj(w, "bomb",hx,hy);
        }
    }
#endif

}
void tick(World * w)
{
    int i,j;
    for(i=0; i<1000; i++)
    {
        Object *obj = &(w->objs[i]);
        if(obj->exist == 0)
            continue;
        if(obj->tick_fn !=0 )
            obj->tick_fn(obj);
        for(j = 0; j<1000; j++)
        {
            Object * obj2 = &(w->objs[j]);
            if(obj2->exist == 0)
                continue;
            if(obj == obj2)
                continue;
            else if(obj->x == obj2->x &&
                    obj->y == obj2->y)
            {

                if(obj->overlap_fn!=0 )
                    obj->overlap_fn(obj,obj2);
            }
        }
    }
}
Cell get_cell(World * w, int x, int y)
{
    Cell c;
    c.bg_color = 0;
    c.fg_color = 0;
    strcpy(c.str," ");
    int i;
    for(i = 0; i<1000; i++)
    {
        Object o = w->objs[i];
        if(o.shape.str == '.')
            continue;
        if(o.exist == 0)
            continue;
        if(o.x == x &&o.y == y)
            return o.shape;
    }
    if(x<0 || y<0 || x >=100 || y >= 100)
        return c;
    return w->map[y][x];
}
void gotoxy(int x, int y)
{
    HANDLE h = GetStdHandle(STD_OUTPUT_HANDLE);
    COORD p = {x,y};
    SetConsoleCursorPosition(h,p);
}
void set_color(char bg, char fg)
{
    WORD c = (bg << 4) | fg;
    HANDLE h = GetStdHandle(STD_OUTPUT_HANDLE);
    SetConsoleTextAttribute(h,c);
}

void draw(World * w)
{
    int cx, cy;
    Cell buffer[20][60];
    memset(buffer,0,sizeof(Cell)*20*20);
    //screen size == 20x20
    for(cy = 0; cy<20; cy++)
    {
        for(cx = 0; cx<60; cx++)
        {
            int hx = w->hero->x;
            int hy = w->hero->y;
            //check error
            //get_cell returns tile or obj shape
            buffer[cy][cx] = get_cell(w,hx-10+cx,hy-10+cy);
        }
    }
    //draw to buffer done

    //draw to screen start
    gotoxy(0,0);
    for(cy = 0; cy<20; cy++)
    {
        for(cx = 0; cx<60; cx++)
        {
            if( buffer[cy][cx].str == '.')
            {
                continue;
            }

            set_color(buffer[cy][cx].bg_color, buffer[cy][cx].fg_color);
            printf(buffer[cy][cx].str);
        }
        printf("\n");
    }
}
void game_over()
{
    system("cls");
    gotoxy(10,10);
    printf("game over..");
    Sleep(3000);
    exit(0);
}
int main()
{
    int cx,cy;
    int i;
    srand(time(0));
    World w;
    memset(&w,0,sizeof(World));

    new_obj(&w,"hero",0,0);
    for(cy = 0;cy<100;cy++)
    {
        for(cx = 0; cx<100;cx++)
        {
            w.map[cy][cx].bg_color = 0;
            w.map[cy][cx].fg_color = 15;
            w.map[cy][cx].layer = 0;
            strcpy(w.map[cy][cx].str,".");
        }
    }
    for(i = 0; i< 400; i++)
    {
        Cell c;
        c.fg_color = 12;
        c.bg_color = 0;
        c.layer  =1;
        strcpy(c.str,"#");
        w.map[rand()%100][rand()%100] = c;
    }
    int k = 50;
    while(k--)
        new_obj(&w,"enemy",rand()%100,rand()%100);
    while(1)
    {
        key_process(&w);
        tick(&w);
        draw(&w);
        Sleep(10);
        if(w.hero->exist == 0)
            game_over();
    }

    return 0;
}
